from http import HTTPStatus
from typing import Any, Literal, TypeAlias, TypeVar, cast

import opentelemetry.context as context
import orjson
from latch_data_validation.data_validation import DataValidationError, validate
from latch_o11y.o11y import trace_function, trace_function_with_span
from opentelemetry.trace.span import Span

from ..asgi_iface import (
    HTTPReceiveCallable,
    HTTPResponseBodyEvent,
    HTTPResponseStartEvent,
    HTTPSendCallable,
)
from .common import Headers, tracer

T = TypeVar("T")

HTTPMethod: TypeAlias = (
    Literal["GET"]
    | Literal["HEAD"]
    | Literal["POST"]
    | Literal["PUT"]
    | Literal["DELETE"]
    | Literal["CONNECT"]
    | Literal["OPTIONS"]
    | Literal["TRACE"]
    | Literal["PATCH"]
)

# >>> O11y

http_request_span_key = context.create_key("http_request_span")


def current_http_request_span() -> Span:
    return cast(Span, context.get_value(http_request_span_key))


# >>> Error classes


class HTTPErrorResponse(RuntimeError):
    def __init__(self, status: HTTPStatus, data: Any, *, headers: Headers = {}):
        self.status = status
        self.data = data
        self.headers = headers


class HTTPInternalServerError(HTTPErrorResponse):
    def __init__(self, data: Any, *, headers: Headers = {}):
        super().__init__(HTTPStatus.INTERNAL_SERVER_ERROR, data, headers=headers)


class HTTPBadRequest(HTTPErrorResponse):
    def __init__(self, data: Any, *, headers: Headers = {}):
        super().__init__(HTTPStatus.BAD_REQUEST, data, headers=headers)


class HTTPForbidden(HTTPErrorResponse):
    def __init__(self, data: Any, *, headers: Headers = {}):
        super().__init__(HTTPStatus.FORBIDDEN, data, headers=headers)


class HTTPConnectionClosedError(RuntimeError): ...


# >>> I/O

# todo(maximsmol): add max body length limit by default


async def receive_class_ext(
    receive: HTTPReceiveCallable, cls: type[T]
) -> tuple[Any, T]:
    data = await receive_json(receive)

    try:
        return data, validate(data, cls)
    except DataValidationError as e:
        raise HTTPBadRequest(e.json()) from None


@trace_function(tracer)
async def receive_class(receive: HTTPReceiveCallable, cls: type[T]) -> T:
    return (await receive_class_ext(receive, cls))[1]


@trace_function(tracer)
async def receive_json(receive: HTTPReceiveCallable) -> Any:
    return orjson.loads(await receive_data(receive))


async def receive_data(receive: HTTPReceiveCallable):
    res = b""
    more_body = True
    while more_body:
        with tracer.start_as_current_span("read chunk") as s:
            msg = await receive()
            if msg.type == "http.disconnect":
                raise HTTPConnectionClosedError()

            res += msg.body
            more_body = msg.more_body

            s.set_attributes({"size": len(msg.body), "more_body": more_body})

    # todo(maximsmol): accumulate instead of overriding
    # todo(maximsmol): probably use the content-length header if present?
    current_http_request_span().set_attribute("http.request_content_length", len(res))

    return res


@trace_function_with_span(tracer)
async def send_http_data(
    s: Span,
    send: HTTPSendCallable,
    status: HTTPStatus,
    data: str | bytes,
    /,
    *,
    content_type: str | bytes | None = "text/plain",
    headers: Headers = {},
):
    if isinstance(data, str):
        data = data.encode("utf-8")

    s.set_attribute("size", len(data))
    headers_to_send: list[tuple[bytes, bytes]] = [
        (b"Content-Length", str(len(data)).encode("latin-1"))
    ]
    for k, v in headers.items():
        if isinstance(k, str):
            k = k.encode("latin-1")
        if isinstance(v, str):
            v = v.encode("latin-1")
        headers_to_send.append((k, v))

    if content_type is not None:
        if isinstance(content_type, str):
            content_type = content_type.encode("latin-1")
        headers_to_send.append((b"Content-Type", content_type))

    await send(
        HTTPResponseStartEvent(
            type="http.response.start", status=status, headers=headers_to_send
        )
    )
    await send(
        HTTPResponseBodyEvent(type="http.response.body", body=data, more_body=False)
    )

    current_http_request_span().set_attribute("http.response_content_length", len(data))


@trace_function(tracer)
async def send_json(
    send: HTTPSendCallable,
    status: HTTPStatus,
    data: Any,
    /,
    *,
    content_type: str = "application/json",
    headers: Headers = {},
):
    return await send_http_data(
        send, status, orjson.dumps(data), content_type=content_type, headers=headers
    )


@trace_function(tracer)
async def send_auto(
    send: HTTPSendCallable,
    status: HTTPStatus,
    data: str | bytes | Any,
    /,
    *,
    headers: Headers = {},
):
    if isinstance(data, str) or isinstance(data, bytes):
        return await send_http_data(send, status, data, headers=headers)

    return await send_json(send, status, data, headers=headers)



from dataclasses import dataclass
from typing import Awaitable, Callable, TypeAlias, TypeVar

from latch_o11y.o11y import AttributesDict, dict_to_attrs

from ..asgi_iface import WebsocketReceiveCallable, WebsocketScope, WebsocketSendCallable
from ..framework.websocket import WebsocketStatus, current_websocket_request_span
from . import common

T = TypeVar("T")


@dataclass
class Context(
    common.Context[WebsocketScope, WebsocketReceiveCallable, WebsocketSendCallable]
):
    def __post_init__(self):
        super().__post_init__()

        if self.auth.oauth_sub is not None:
            current_websocket_request_span().set_attribute(
                "enduser.id", self.auth.oauth_sub
            )

    def add_request_span_attrs(self, data: AttributesDict, prefix: str):
        current_websocket_request_span().set_attributes(dict_to_attrs(data, prefix))


HandlerResult = str | tuple[WebsocketStatus, str]
Handler: TypeAlias = Callable[
    [Context],
    Awaitable[HandlerResult],
]
Route: TypeAlias = Handler



from enum import Enum
from typing import Any, TypeVar, cast

import opentelemetry.context as context
from latch_data_validation.data_validation import DataValidationError, validate
from latch_o11y.o11y import trace_function, trace_function_with_span
from opentelemetry.trace.span import Span
import orjson

from ..asgi_iface import (
    WebsocketAcceptEvent,
    WebsocketCloseEvent,
    WebsocketReceiveCallable,
    WebsocketSendCallable,
    WebsocketSendEvent,
)
from .common import Headers, tracer

T = TypeVar("T")

websocket_request_span_key = context.create_key("websocket_request_span")


def current_websocket_request_span() -> Span:
    return cast(Span, context.get_value(websocket_request_span_key))


# >>> Error classes


class WebsocketStatus(int, Enum):
    """
    https://www.rfc-editor.org/rfc/rfc6455.html#section-7.4.1
    """

    normal = 1000
    """
    1000 indicates a normal closure, meaning that the purpose for
    which the connection was established has been fulfilled.
    """

    going_away = 1001
    """
    1001 indicates that an endpoint is "going away", such as a server
    going down or a browser having navigated away from a page.
    """

    protocol_error = 1002
    """
    1002 indicates that an endpoint is terminating the connection due
    to a protocol error.
    """

    unsupported = 1003
    """
    1003 indicates that an endpoint is terminating the connection
    because it has received a type of data it cannot accept (e.g., an
    endpoint that understands only text data MAY send this if it
    receives a binary message).
    """

    reserved = 1004
    """
    Reserved.  The specific meaning might be defined in the future.
    """

    no_status = 1005
    """
    1005 is a reserved value and MUST NOT be set as a status code in a
    Close control frame by an endpoint.  It is designated for use in
    applications expecting a status code to indicate that no status
    code was actually present.
    """

    abnormal = 1006
    """
    1006 is a reserved value and MUST NOT be set as a status code in a
    Close control frame by an endpoint.  It is designated for use in
    applications expecting a status code to indicate that the
    connection was closed abnormally, e.g., without sending or
    receiving a Close control frame.
    """

    unsupported_payload = 1007
    """
    1007 indicates that an endpoint is terminating the connection
    because it has received data within a message that was not
    consistent with the type of the message (e.g., non-UTF-8 [RFC3629]
    data within a text message).
    """

    policy_violation = 1008
    """
    1008 indicates that an endpoint is terminating the connection
    because it has received a message that violates its policy.  This
    is a generic status code that can be returned when there is no
    other more suitable status code (e.g., 1003 or 1009) or if there
    is a need to hide specific details about the policy.
    """

    too_large = 1009
    """
    1009 indicates that an endpoint is terminating the connection
    because it has received a message that is too big for it to
    process.
    """

    mandatory_extension = 1010
    """
    1010 indicates that an endpoint (client) is terminating the
    connection because it has expected the server to negotiate one or
    more extension, but the server didn't return them in the response
    message of the WebSocket handshake.  The list of extensions that
    are needed SHOULD appear in the /reason/ part of the Close frame.
    Note that this status code is not used by the server, because it
    can fail the WebSocket handshake instead.
    """

    server_error = 1011
    """
    1011 indicates that a server is terminating the connection because
    it encountered an unexpected condition that prevented it from
    fulfilling the request.
    """

    tls_handshake_fail = 1015
    """
    1015 is a reserved value and MUST NOT be set as a status code in a
    Close control frame by an endpoint.  It is designated for use in
    applications expecting a status code to indicate that the
    connection was closed due to a failure to perform a TLS handshake
    (e.g., the server certificate can't be verified).
    """


class WebsocketErrorResponse(RuntimeError):
    def __init__(self, status: WebsocketStatus, data: Any, *, headers: Headers = {}):
        super().__init__()
        self.status = status
        self.data = data
        self.headers = headers


class WebsocketBadMessage(WebsocketErrorResponse):
    def __init__(self, data: Any, *, headers: Headers = {}):
        super().__init__(WebsocketStatus.policy_violation, data, headers=headers)


class WebsocketInternalServerError(WebsocketErrorResponse):
    def __init__(self, data: Any, *, headers: Headers = {}):
        super().__init__(WebsocketStatus.server_error, data, headers=headers)


class WebsocketConnectionClosedError(RuntimeError): ...


# >>> I/O


async def receive_websocket_data(receive: WebsocketReceiveCallable):
    with tracer.start_as_current_span("read websocket message") as s:
        msg = await receive()

        s.set_attribute("type", msg.type)

        if msg.type == "websocket.connect":
            # todo(ayush): allow upgrades here as well?
            raise WebsocketBadMessage("connection has already been established")

        if msg.type == "websocket.disconnect":
            raise WebsocketConnectionClosedError()

        if msg.bytes is not None:
            res = msg.bytes
        elif msg.text is not None:
            res = msg.text.encode("utf-8")
        else:
            raise WebsocketBadMessage("empty message")

        s.set_attribute("size", len(res))
        return res


@trace_function(tracer)
async def receive_websocket_json(receive: WebsocketReceiveCallable) -> Any:
    orjson.loads(await receive_websocket_data(receive))


@trace_function(tracer)
async def receive_websocket_class_ext(
    receive: WebsocketReceiveCallable, cls: type[T]
) -> tuple[Any, T]:
    data = await receive_websocket_json(receive)

    try:
        return data, validate(data, cls)
    except DataValidationError as e:
        raise WebsocketBadMessage(e.json()) from None


@trace_function(tracer)
async def receive_websocket_class(receive: WebsocketReceiveCallable, cls: type[T]) -> T:
    return (await receive_websocket_class_ext(receive, cls))[1]


@trace_function_with_span(tracer)
async def send_websocket_data(
    s: Span,
    send: WebsocketSendCallable,
    data: str | bytes,
):
    if isinstance(data, str):
        data = data.encode("utf-8")

    s.set_attribute("body.size", len(data))

    await send(WebsocketSendEvent(type="websocket.send", bytes=data, text=None))

    current_websocket_request_span().set_attribute(
        "websocket.sent_message_content_length", len(data)
    )


@trace_function(tracer)
async def accept_websocket_connection(
    send: WebsocketSendCallable,
    receive: WebsocketReceiveCallable,
    /,
    *,
    subprotocol: str | None = None,
    headers: Headers = {},
):
    msg = await receive()
    if msg.type != "websocket.connect":
        raise WebsocketBadMessage("cannot accept connection without connection request")

    headers_to_send: list[tuple[bytes, bytes]] = []

    for k, v in headers.items():
        if isinstance(k, str):
            k = k.encode("latin-1")
        if isinstance(v, str):
            v = v.encode("latin-1")
        headers_to_send.append((k, v))

    await send(
        WebsocketAcceptEvent(
            type="websocket.accept", subprotocol=subprotocol, headers=headers_to_send
        )
    )


@trace_function_with_span(tracer)
async def close_websocket_connection(
    s: Span,
    send: WebsocketSendCallable,
    /,
    *,
    status: WebsocketStatus,
    data: str,
):
    s.set_attribute("reason", data)

    await send(WebsocketCloseEvent("websocket.close", status.value, data))

    current_websocket_request_span().set_attribute("websocket.http.close_reason", data)



import asyncio
import dataclasses
import traceback
from http import HTTPStatus
from typing import Awaitable, get_args

import opentelemetry.context as context
from hypercorn.typing import ASGIReceiveCallable, ASGISendCallable, Scope
from latch_data_validation.data_validation import untraced_validate, validate
from latch_o11y.o11y import log
from opentelemetry.propagate import set_global_textmap
from opentelemetry.trace import get_current_span, get_tracer

from .asgi_iface import (
    HTTPDisconnectEvent,
    HTTPReceiveCallable,
    HTTPRequestEvent,
    HTTPScope,
    HTTPSendCallable,
    HTTPSendEvent,
    LifespanReceiveCallable,
    LifespanScope,
    LifespanSendCallable,
    LifespanSendEvent,
    LifespanShutdownCompleteEvent,
    LifespanShutdownEvent,
    LifespanShutdownFailedEvent,
    LifespanStartupCompleteEvent,
    LifespanStartupEvent,
    LifespanStartupFailedEvent,
    WebsocketReceiveCallable,
    WebsocketReceiveEventT,
    WebsocketScope,
    WebsocketSendCallable,
    WebsocketSendEventT,
    type_str,
)
from .context import http, websocket
from .datadog_propagator import DDTraceContextTextMapPropagator
from .framework.http import (
    HTTPErrorResponse,
    HTTPInternalServerError,
    current_http_request_span,
    http_request_span_key,
    send_auto,
    send_http_data,
)
from .framework.websocket import (
    WebsocketErrorResponse,
    WebsocketInternalServerError,
    WebsocketStatus,
    accept_websocket_connection,
    close_websocket_connection,
    current_websocket_request_span,
    websocket_request_span_key,
)

tracer = get_tracer(__name__)


# todo(maximsmol): ASGI instrumentation should trace lifespan


class LatchASGIServer:
    http_routes: dict[str, http.Route]
    websocket_routes: dict[str, websocket.Route]
    startup_tasks: list[Awaitable] = []
    shutdown_tasks: list[Awaitable] = []

    def __init__(
        self,
        http_routes: dict[str, http.Route],
        websocket_routes: dict[str, websocket.Route],
        startup_tasks: list[Awaitable] = [],
        shutdown_tasks: list[Awaitable] = [],
    ):
        self.http_routes = http_routes
        self.websocket_routes = websocket_routes
        self.startup_tasks = startup_tasks
        self.shutdown_tasks = shutdown_tasks

    async def scope_lifespan(
        self,
        scope: LifespanScope,
        receive: LifespanReceiveCallable,
        send: LifespanSendCallable,
    ):
        await log.info(
            f"Waiting for lifespan events (ASGI v{scope.asgi.version} @ spec"
            f" v{scope.asgi.spec_version})"
        )
        while True:
            message = await receive()
            await log.info(repr(message.type))

            if isinstance(message, LifespanStartupEvent):
                with tracer.start_as_current_span("startup"):
                    try:
                        await log.info("Executing startup tasks")
                        # todo(maximsmol): debug clock skew on connection reset
                        await asyncio.gather(*self.startup_tasks)

                        with tracer.start_as_current_span("send completion event"):
                            await send(
                                LifespanStartupCompleteEvent(
                                    "lifespan.startup.complete"
                                )
                            )
                    except Exception as e:
                        with tracer.start_as_current_span("send failure event"):
                            await send(
                                LifespanStartupFailedEvent(
                                    "lifespan.startup.failed", str(e)
                                )
                            )

                        raise e
            elif isinstance(message, LifespanShutdownEvent):
                with tracer.start_as_current_span("shutdown"):
                    try:
                        await asyncio.gather(*self.shutdown_tasks)

                        with tracer.start_as_current_span("send completion event"):
                            await send(
                                LifespanShutdownCompleteEvent(
                                    "lifespan.shutdown.complete"
                                )
                            )
                    except Exception as e:
                        with tracer.start_as_current_span("send failure event"):
                            await send(
                                LifespanShutdownFailedEvent(
                                    "lifespan.shutdown.failed", str(e)
                                )
                            )

                        raise e

                    break

    async def scope_websocket(
        self,
        scope: WebsocketScope,
        receive: WebsocketReceiveCallable,
        send: WebsocketSendCallable,
    ):
        ctx_reset_token: object | None = None
        try:
            new_ctx = context.set_value(websocket_request_span_key, get_current_span())
            ctx_reset_token = context.attach(new_ctx)

            current_websocket_request_span().set_attribute("resource.name", scope.path)

            handler = self.websocket_routes.get(scope.path)
            if handler is None:
                msg = f"Websocket {scope.path} not found"

                await log.info(msg)
                await close_websocket_connection(
                    send, status=WebsocketStatus.policy_violation, data=msg
                )
                return

            await log.info(f"Websocket {scope.path}")

            try:
                try:
                    ctx = websocket.Context(scope, receive, send)

                    await accept_websocket_connection(ctx.send, ctx.receive)
                    res = await handler(ctx)

                    if isinstance(res, tuple):
                        status, data = res
                    else:
                        status = WebsocketStatus.normal
                        data = res

                except WebsocketErrorResponse as e:
                    raise e
                except Exception as e:
                    raise WebsocketInternalServerError(str(e)) from e
            except WebsocketErrorResponse as e:
                await close_websocket_connection(
                    send, status=WebsocketStatus.server_error, data=str(e.data)
                )

                if e.status == HTTPStatus.INTERNAL_SERVER_ERROR:
                    traceback.print_exc()
            else:
                await close_websocket_connection(send, status=status, data=data)
        finally:
            if ctx_reset_token is not None:
                context.detach(ctx_reset_token)

    async def scope_http(
        self, scope: HTTPScope, receive: HTTPReceiveCallable, send: HTTPSendCallable
    ):
        ctx_reset_token: object | None = None
        try:
            new_ctx = context.set_value(http_request_span_key, get_current_span())
            ctx_reset_token = context.attach(new_ctx)

            current_http_request_span().set_attribute("resource.name", scope.path)
            await log.info(f"{scope.method} {scope.path}")

            with tracer.start_as_current_span("find route handler"):
                route = self.http_routes.get(scope.path)

            if not isinstance(route, tuple):
                methods = ["POST"]
                handler = route
            else:
                methods, handler = route

            if scope.method not in methods:
                if len(methods) == 1:
                    methods_str = methods[0]
                elif len(methods) == 2:
                    methods_str = f"{methods[0]} and {methods[1]}"
                else:
                    methods_str = ", and ".join([", ".join(methods[:-1]), methods[-1]])

                await send_http_data(
                    send,
                    HTTPStatus.METHOD_NOT_ALLOWED,
                    f"Only {methods_str} requests are supported",
                )
                return

            if handler is None:
                # todo(maximsmol): better error message
                await log.info("Not found")
                await send_http_data(send, HTTPStatus.NOT_FOUND, "Not found")
                return

            try:
                try:
                    ctx = http.Context(scope, receive, send)
                    res = await handler(ctx)

                    if res is not None:
                        with tracer.start_as_current_span("send response"):
                            await send_auto(send, HTTPStatus.OK, res)
                    return
                except HTTPErrorResponse as e:
                    raise e
                except Exception as e:
                    # todo(maximsmol): better error message
                    raise HTTPInternalServerError("Internal error") from e
            except HTTPErrorResponse as e:
                await send_auto(send, e.status, {"error": e.data}, headers=e.headers)
                if e.status == HTTPStatus.INTERNAL_SERVER_ERROR:
                    # await log.exception() # fixme(maximsmol)
                    traceback.print_exc()
                return
        finally:
            if ctx_reset_token is not None:
                context.detach(ctx_reset_token)

    async def raw_app(
        self, scope: Scope, receive: ASGIReceiveCallable, send: ASGISendCallable
    ):
        try:
            if scope["type"] == "lifespan":

                async def ls_receive():
                    x = await receive()

                    if x["type"] == type_str(LifespanStartupEvent):
                        return untraced_validate(x, LifespanStartupEvent)

                    if x["type"] == type_str(LifespanShutdownEvent):
                        return untraced_validate(x, LifespanShutdownEvent)

                    raise RuntimeError(
                        f"unknown lifespan event type: {repr(x['type'])}"
                    )

                async def ls_send(e: LifespanSendEvent):
                    data = dataclasses.asdict(e)
                    await send(data)

                return await self.scope_lifespan(
                    untraced_validate(scope, LifespanScope), ls_receive, ls_send
                )

            if scope["type"] == "websocket":

                async def ws_receive():
                    x = await receive()

                    for e in get_args(WebsocketReceiveEventT):
                        if x["type"] != type_str(e):
                            continue

                        return untraced_validate(x, e)

                    raise RuntimeError(
                        f"unknown websocket event type: {repr(x['type'])}"
                    )

                async def ws_send(e: WebsocketSendEventT):
                    data = dataclasses.asdict(e)
                    await send(data)

                return await self.scope_websocket(
                    untraced_validate(scope, WebsocketScope), ws_receive, ws_send
                )

            if scope["type"] == "http":

                async def http_receive():
                    x = await receive()

                    if x["type"] == type_str(HTTPRequestEvent):
                        return validate(x, HTTPRequestEvent)

                    if x["type"] == type_str(HTTPDisconnectEvent):
                        return validate(x, HTTPDisconnectEvent)

                    raise RuntimeError(f"unknown http event type: {repr(x['type'])}")

                async def http_send(e: HTTPSendEvent):
                    data = dataclasses.asdict(e)
                    await send(data)

                return await self.scope_http(
                    validate(scope, HTTPScope), http_receive, http_send
                )

            raise RuntimeError(f"unsupported protocol: {repr(scope['type'])}")
        except Exception as e:
            await log.exception("Fallback exception handler:")
            raise e


set_global_textmap(DDTraceContextTextMapPropagator())



from dataclasses import dataclass
from typing import Any, Awaitable, Callable, TypeAlias, TypeVar

from latch_o11y.o11y import AttributesDict, dict_to_attrs, trace_app_function

from ..asgi_iface import HTTPReceiveCallable, HTTPScope, HTTPSendCallable
from ..framework.http import HTTPMethod, current_http_request_span, receive_class_ext
from . import common

T = TypeVar("T")


@dataclass
class Context(common.Context[HTTPScope, HTTPReceiveCallable, HTTPSendCallable]):
    def __post_init__(self):
        super().__post_init__()

        if self.auth.oauth_sub is not None:
            current_http_request_span().set_attribute("enduser.id", self.auth.oauth_sub)

    def add_request_span_attrs(self, data: AttributesDict, prefix: str):
        current_http_request_span().set_attributes(dict_to_attrs(data, prefix))

    @trace_app_function
    async def receive_request_payload(self, cls: type[T]) -> T:
        json, res = await receive_class_ext(self.receive, cls)

        # todo(maximsmol): datadog has shit support for events
        # current_http_request_span().add_event(
        #     "request payload", dict_to_attrs(json, "data")
        # )
        self.add_request_span_attrs(json, "http.request_payload")

        return res


HandlerResult: TypeAlias = Any | None
Handler: TypeAlias = Callable[
    [Context],
    Awaitable[HandlerResult],
]
Route: TypeAlias = Handler | tuple[list[HTTPMethod], Handler]



from opentelemetry import trace
from opentelemetry.context.context import Context
from opentelemetry.propagators import textmap


class DDTraceContextTextMapPropagator(textmap.TextMapPropagator):
    # https://github.com/open-telemetry/opentelemetry-python-contrib/blob/934af7ea4f9b1e0294ced6a014d6eefdda156b2b/exporter/opentelemetry-exporter-datadog/src/opentelemetry/exporter/datadog/propagator.py
    def extract(
        self,
        carrier: textmap.CarrierT,
        context: Context | None = None,
        getter: textmap.Getter[textmap.CarrierT] = textmap.default_getter,
    ) -> Context:
        # todo(maximsmol): warn on duplicate headers
        # todo(maximsmol): ignore invalid headers

        if context is None:
            context = Context()

        trace_id = getter.get(carrier, "x-datadog-trace-id")
        if trace_id is None or len(trace_id) == 0:
            return context

        parent_id = getter.get(carrier, "x-datadog-parent-id")
        if parent_id is None or len(parent_id) == 0:
            return context

        sampling_priority = getter.get(carrier, "x-datadog-sampling-priority")

        trace_flags = trace.TraceFlags()
        if sampling_priority is not None and len(sampling_priority) > 0:
            if int(sampling_priority[0]) in (
                1,  # auto keep
                2,  # user keep
            ):
                trace_flags = trace.TraceFlags(trace.TraceFlags.SAMPLED)

        dd_origin = getter.get(carrier, "x-datadog-origin")

        trace_state = []
        if dd_origin is not None and len(dd_origin) > 0:
            trace_state.append(("dd_origin", dd_origin[0]))

        span_context = trace.SpanContext(
            trace_id=int(trace_id[0]),
            span_id=int(parent_id[0]),
            is_remote=True,
            trace_flags=trace_flags,
            trace_state=trace.TraceState(trace_state),
        )
        return trace.set_span_in_context(trace.NonRecordingSpan(span_context), context)

    def inject(
        self,
        carrier: textmap.CarrierT,
        context: Context | None = None,
        setter: textmap.Setter[textmap.CarrierT] = textmap.default_setter,
    ) -> None:
        span = trace.get_current_span(context)
        if span == trace.INVALID_SPAN:
            return

        span_context = span.get_span_context()
        if span_context == trace.INVALID_SPAN_CONTEXT:
            return

        setter.set(
            carrier,
            "x-datadog-trace-id",
            str(span_context.trace_id & 0xFFFF_FFFF_FFFF_FFFF),
        )
        setter.set(
            carrier,
            "x-datadog-parent-id",
            str(span_context.span_id),
        )
        setter.set(
            carrier,
            "x-datadog-sampling-priority",
            "0"  # auto reject
            if span_context.trace_flags & trace.TraceFlags.SAMPLED == 0
            else "1",  # auto keep
        )
        if "dd_origin" in span_context.trace_state:
            setter.set(
                carrier, "x-datadog-origin", span_context.trace_state["dd_origin"]
            )

    @property
    def fields(self) -> set[str]:
        return {
            "x-datadog-trace-id",
            "x-datadog-parent-id",
            "x-datadog-sampling-priority",
            "x-datadog-origin",
        }
